#!/usr/bin/python
#
# written RGC 2003 Jul 31
# revised KMB 2003 Aug 01
# revised KMB 2003 Aug 13 
# revised RGC 2003 Oct 08
# revised RGC 2003 Oct 22
# revised RGC 2004 May 11 (removed spurious linebreaks in file)
# python parallel pairwise pinging daemon
# performs pinging tasks as specified in a file
# loops around pinging

import time, sched, commands, string, os, sys, re, calendar, socket


# compares IP addresses
def sitesmatch(site1, site2):

  return site1 == site2
  
def launch_it(command, filename, rtime):
  now = time.time()
  # TODO  could we find a good way to grab CPU load here?
  outputf = os.popen(command)
  output  = outputf.read()
  try:
    fileh = open(filename,'a')
  except IOError:
    print >> sys.stderr, 'Unable to write to file',filename
    sys.exit()
  try:
    f = open('/proc/loadavg', 'r')
    u_avg = f.read().split()            # collect the load average
    f.close()
  except IOError:
    u_avg = (-1,-1)   # didn't get load average so pretend it was 0
  print >> fileh, rtime, time.strftime(dateoutformat,time.localtime(rtime)),
  print >> fileh, u_avg[0],
  for line in output.split('\n'):
    ip = ipregexp.search(line)
    if ip:
      iptime = timeregexp.search(line)
      if iptime:
        print >> fileh, ip.group(0), iptime.group(0),
      else:
        print >> fileh, ip.group(0), "0ms",
    else:
      print >> fileh, "0.0.0.0 0ms",
  print >> fileh
  fileh.close()

# main program start...


dateoutformat = '%Y_%02m_%02d_%02H_%02M'
dateinformat  = '%Y_%m_%d_%H_%M'
ipregexp   = re.compile(r'(\d\d*\.\d\d*\.\d\d*\.\d\d*)')
timeregexp = re.compile(r'(\d*\.+\d+ms)')
nargs = len(sys.argv)

# get our configuration file
if nargs == 1: filename = 'ppppd.conf'
else: filename = sys.argv[1]
myname = socket.gethostname()
try: 
  myip= socket.gethostbyname(myname)
except socket.gaierror:
  print >> sys.stderr >> "This host ",myname," does not resolve to an IP address "
  sys.exit()
print 'ppppd.py running on',myname
print 'Reading configuration from', filename

try:
  file = open(filename,'r')
except IOError:
  print >> sys.stderr, 'Unable to read from file', filename
  print >> sys.stderr, 'Usage: ppppd.py [conf.filename]'
  sys.exit()

lineno = 1
stime = None
etime = None
sites = None
pingcomm = None
instant = 0
delay = 60
site1 = []
site2 = []
sitefile = []

for line in file.readlines():
  if not line: break
  if re.match('^#',line):    # skip comments
    lineno += 1
    continue
  # remove trailing newlines
  if line.endswith('\n'): line=line[:-1]
  # get the start time for the run
  if re.match('^(?i)START ',line):
    line = re.sub('^(?i)START ','',line)
    try:
      stime  = time.strptime(line, dateinformat)
      stuple = -1,
      stime  = stime[0:8] + stuple # set dst to unknown
      stime  = time.mktime(stime)
    except ValueError:
      print >> sys.stderr, 'Start time incorrect at line', lineno, line
      print >> sys.stderr, 'Example date format: 1971_1_1_22_00'
      sys.exit()
  elif re.match('^(?i)NOW', line):
      instant = 1
  elif re.match('^(?i)END ',line):
    line= re.sub('^(?i)END ','',line)
    try:
      #  etime= time.strptime(line, '%d %b %Y %H:%M')
      etime  = time.strptime(line, dateinformat)
      stuple = -1,
      etime  = etime[0:8] + stuple   # set dst to unknown
      etime  = time.mktime(etime)
    except ValueError:
      print >> sys.stderr, 'End time incorrect at line',lineno,line
      print >> sys.stderr, 'Example date format: 1971_1_1_22_00'
      sys.exit(1)
  elif re.match('^(?i)DELAY ',line):
    line = re.sub('^(?i)DELAY ','',line)
    try:
      delay = string.atoi(line)
    except ValueError:
      print >> sys.stderr, 'No value specified for delay at line', lineno, line
      sys.exit(2)
  elif re.match('^(?i)SITES ',line):
    parse = line.split()
    if len(parse) !=4:
      print >> sys.stderr, 'Error in sites line at line',lineno,line
      print >> sys.stderr,  'Format should be: site1  site2   outputfile'
      sys.exit(3)
    site1.append(parse[1])
    site2.append(parse[2])
    sitefile.append(parse[3])
  elif re.match('^(?i)PROBE ',line):
    pingcomm = re.sub('^(?i)PROBE ','',line).strip()
    if len(pingcomm) == 0:
      print >> sys.stderr, 'Probe command not defined at line',lineno, line
      sys.exit()
# check here if the command actually exists and is runnable
  else:
    print >>  sys.stderr, 'Unknown line at line', lineno, line
    sys.exit()
  lineno += 1

nsites = len(site1)
file.close()
now = time.time()

if instant == 1:
  print 'NOW command is overriding start and end times.'
  stime = now+30
  etime = stime+(2*delay*nsites)+1  # enough for two complete cycles
if stime==None:
  print >> sys.stderr,  'Start line missing from config file.  Please provide start time.'
  sys.exit(5)
if etime==None:
  print >> sys.stderr, 'End line missing from config file.  Please provide end time.'
  sys.exit(4)
if int(now)+20>stime:  # allow some leeway for slow execution
  print >> sys.stderr, 'Missed the start time. Starting in 20 seconds.'
  sys.exit(6)
if etime < stime:
  print >> sys.stderr, 'End time must be after start time.'
  sys.exit(7)
print  'Data collection start time :', time.asctime(time.localtime(stime))
print  'Data collection end time   :', time.asctime(time.localtime(etime))
print  'Data collection every',delay,'seconds'

for i in range(nsites):
  print  'Pings from',site1[i],'to',site2[i],'saved in',sitefile[i]
# check if files exist
fileerror= 0
sitefound= 0
for i in range(nsites):
#Convert site names into site IP addresses
  strip= site1[i]+':'
  strip= strip[:strip.index(':')]
  try:
    sip= socket.gethostbyname(strip)
  except socket.gaierror:
    print >> sys.stderr >> strip, " does not resolve to an IP address "
    sys.exit()
  site1[i]= sip
  strip= site2[i]+':'
  strip= strip[:strip.index(':')]
  try:
    sip= socket.gethostbyname(strip)
  except socket.gaierror:
    print >> sys.stderr >> strip, " does not resolve to an IP address "
    sys.exit()
  site2[i]= sip
  if sitesmatch(site1[i],myip) or sitesmatch(site2[i],myip):
    sitefound= 1
  try:
    file= open(sitefile[i],'r')
  except (IOError):
    pass
  else:
    fileerror = 1
    print >> sys.stderr, 'Please delete',sitefile[i], ' before running'

if not sitefound:
  print >> sys.stderr, 'This site',myname,'is not one of the sites to ping from'
  sys.exit(8)

if pingcomm == None:
  print >> sys.stderr, 'No probe command defined.'
  print >> sys.stderr, 'Use probe [command] in configuration file.'
  sys.exit(9)

if fileerror == 1:
  sys.exit(10)

# check if files can be written
for i in range(nsites):
  try:
    file = open(sitefile[i],'w')
  except (IOError):
    print >> sys.stderr, 'Cannot write to ',sitefile[i]
    sys.exit(11)
  file.close()

s = sched.scheduler(time.time, time.sleep)
next = stime
print >> sys.stderr, 'Start at',    time.asctime(time.localtime(next))
print >> sys.stderr, 'Time is now', time.asctime(time.localtime(now))
i=0

while not sitesmatch (site1[i],myip) and not sitesmatch (site2[i],myip):
  i += 1
  i %= nsites
  next +=  delay

while next < etime:
  if sitesmatch(site1[i],myip): command = pingcomm+' '+site2[i]
  else: command = pingcomm+' '+site1[i]
  print 'Next command at:', time.asctime(time.localtime(next)),':',command
  s.enterabs(next,1,launch_it,(command, sitefile[i], next))
  s.run()
  next += delay
  i += 1
  i %= nsites
  while not sitesmatch (site1[i],myip) and not sitesmatch (site2[i],myip):
    i += 1
    i %= nsites
    next += delay
